const express = require('express');
const multer = require('multer');
const axios = require('axios');
const csv = require('csv-parser');
const XLSX = require('xlsx');
const fs = require('fs');
const path = require('path');
const { parse } = require('json2csv');

const app = express();
const port = 3001;

const upload = multer({ dest: 'uploads/' });

app.use(express.static('public')); // Para servir archivos estáticos

// Función para hacer un delay
const delay = (ms) => new Promise(resolve => setTimeout(resolve, ms));

app.post('/upload', upload.single('file'), async (req, res) => {
  const filePath = path.join(__dirname, 'uploads', req.file.filename);
  const results = [];
  const promises = [];

  try {
    if (req.file.mimetype === 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet' || req.file.mimetype === 'application/vnd.ms-excel') {
      // Leer archivo Excel
      const workbook = XLSX.readFile(filePath);
      const sheet_name_list = workbook.SheetNames;
      const data = XLSX.utils.sheet_to_json(workbook.Sheets[sheet_name_list[0]]);

      for (const row of data) {
        promises.push(makeRequest(row, results));
      }
    } else if (req.file.mimetype === 'text/csv') {
      // Leer archivo CSV
      fs.createReadStream(filePath)
        .pipe(csv())
        .on('data', (data) => {
          promises.push(makeRequest(data, results));
        })
        .on('end', async () => {
          // Esperar a que todas las solicitudes se completen
          await Promise.all(promises);

          // Crear archivo CSV con los resultados
          const fields = ['pasarela', 'codigoUnico', 'terminalesCreadas', 'codigoRespuesta', 'descripcion'];
          const opts = { fields, delimiter: ';' };
          const csv = parse(results, opts);

          const outputFilePath = path.join(__dirname, 'uploads', 'resultado.csv');
          fs.writeFileSync(outputFilePath, csv);

          // Enviar archivo como respuesta
          res.download(outputFilePath, 'resultado.csv', (err) => {
            if (err) {
              console.error(err);
              res.status(500).send('Error al generar el archivo de resultados.');
            }
            // Limpiar archivos temporales
            fs.unlink(filePath, (err) => { if (err) console.error(err); });
            fs.unlink(outputFilePath, (err) => { if (err) console.error(err); });
          });
        });
    } else {
      return res.status(400).json({ message: 'Formato de archivo no soportado' });
    }
  } catch (error) {
    res.status(500).json({ error: error.message });
  }
});

async function makeRequest(row, results) {
  // Agregar delay antes de hacer la solicitud
  await delay(5000); // Delay de 4000 ms (4 segundo)

  const options = {
    method: 'POST',
    url: 'https://orquestador-pasarelas-gateway-alcance.crediservices.credibanco.com/credibanco/api/pasarelas/orquestador/v1/externo/notificaciones/terminales',
    headers: {
      'Content-Type': 'application/json',
      'Authorization': 'Basic Y3JlZGliYW5jb3BydWViYXM6Q29sb21iaWEyMyo='
    },
    data: {
      "pasarela": row.pasarela,
      "codigoUnico": row.codigoUnico,
      "tipoComercio": "NORMAL",
      "cantidadTerminales": 1
    }
  };

  try {
    const response = await axios(options);
    results.push({
      pasarela: row.pasarela,
      codigoUnico: row.codigoUnico,
      terminalesCreadas: response.data.terminalesCreadas ? response.data.terminalesCreadas.join(', ') : '',
      codigoRespuesta: response.data.codigoRespuesta || '',
      descripcion: response.data.descripcion || ''
    });
    console.log(`Respuesta para ${row.pasarela} - ${row.codigoUnico}:`, response.data);
  } catch (error) {
    results.push({
      pasarela: row.pasarela,
      codigoUnico: row.codigoUnico,
      terminalesCreadas: '',
      codigoRespuesta: '',
      descripcion: `Error: ${error.message}`
    });
    console.error(`Error para ${row.pasarela} - ${row.codigoUnico}:`, error.message);
  }
}

app.listen(port, () => {
  console.log(`Servidor corriendo en http://localhost:${port}`);
});
